#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""utils.py

various utilities for the module

"""
import logging
import pkgutil
import re
from functools import wraps
from timeit import default_timer
from trieregex import TrieRegEx as TRE

# requirements
import numpy as np
from pandas import NA, DataFrame
from unidecode import unidecode

logger = logging.getLogger(__name__)


def time_it(func):
    """
    decorator for printing the execution time of a function call
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = default_timer()
        result = func(*args, **kwargs)
        end = default_timer()
        name = func.__name__
        logger.info(f"{name} took {round(end - start, 2)} seconds")
        return result
    return wrapper


###########
# ANCHORS #
###########
def chunk_anchors(lst, n, exclude={0, 1}):
    """yields n-sized chunks from lst without excluded values"""
    lst_rm = [i for i in lst if i not in exclude]
    for i in range(0, len(lst_rm), n):
        yield lst_rm[i:i + n]


def apply_correction(row, correction):
    """Applies correction to one row in column.

    :param list row: row of dump with [value, context, contextend]
    :param int correction: offset to add/subtract from position
    """

    value, lower_bound, upper_bound = row

    if value == -1:
        # missing value
        return value

    value += correction
    if value < lower_bound:
        value = lower_bound
    if value > upper_bound:
        value = upper_bound
    return value


def correct_anchors(df, corrections):
    """Corrects df columns via dictionary of corrections

    :param DataFrame df: dump
    :param dict corrections: {anchor_name (0-9, column in df): offset (int)}
    """
    for correction in corrections:
        if correction in df.columns:
            logger.info(f'correcting anchor {correction} by offset {corrections[correction]}')
            df[correction] = df[
                [correction, 'context', 'contextend']
            ].apply(lambda x: apply_correction(x, corrections[correction]), axis=1)
        else:
            logger.warning(f'anchor "{correction}" not in dataframe')
    return df


###########
# QUERIES #
###########
CQP_META = {
    ".": r"\.",
    "?": r"\?",
    "*": r"\*",
    "+": r"\+",
    "|": r"\|",
    "(": r"\(",
    ")": r"\)",
    "[": r"\[",
    "]": r"\]",
    "{": r"\{",
    "}": r"\}",
    "^": r"\^",
    "$": r"\$",
    '"': r"\"",
    "'": r"\'"
}


def cqp_escape(token):
    """ escape CQP meta-characters

    :param str token: string to escape
    :return: escaped string
    :rtype: str

    """
    return token.translate(str.maketrans(CQP_META))


def cqp_unescape(token):

    token = token.replace("\\\\", "\\")

    for unescaped, escaped in CQP_META.items():
        token = token.replace(escaped, unescaped)

    return token


def format_cqp_query(items, p_query='word', s_query=None, flags="", escape=True):
    """ wrapper for easy queries

    :return: valid cqp query
    :rtype: str
    """

    # three-way categorisation
    mwu_queries = list()
    single_items = list()
    single_items_escaped = list()

    for item in items:

        tokens = item.split(" ")

        if len(tokens) > 1:
            mwu_query = ""
            for token in tokens:
                token = cqp_escape(token) if escape else token
                mwu_query += f'[{p_query}="{token}"{flags}]'
            mwu_queries.append("(" + mwu_query + ")")

        else:

            if escape:
                item_escaped = cqp_escape(item)
                if item_escaped == item:
                    single_items.append(item)
                else:
                    single_items_escaped.append(item_escaped)
            else:
                item_escaped = cqp_escape(item)
                if item_escaped == item:
                    single_items.append(item)
                else:
                    single_items_escaped.append(item)

    singles_queries = []
    if len(single_items) > 0:
        reg = TRE(*single_items).regex().replace("?:", "")
        singles_queries = ["([" + p_query + "=" + '"' + reg + '"' + flags + "])"]

    singles_queries_escaped = []
    if len(single_items_escaped) > 0:
        reg = "|".join(single_items_escaped)
        singles_queries = ["([" + p_query + "=" + '"' + reg + '"' + flags + "])"]

    queries = mwu_queries + singles_queries + singles_queries_escaped

    # disjunctive join
    query = ' | '.join(queries)

    # add s_query (optional)
    if s_query is not None:
        query = f'({query}) within {s_query};'

    # return
    return query


def find_all(pattern, string):
    p = re.compile(pattern)
    matches = dict()
    for m in p.finditer(string):
        matches[m.group()] = [m.start(), m.end()]
    return matches


def preprocess_query(query):
    """
    - get s_query and strip 'within' statement
    - get anchors
    - get macros
    - get wordlists

    :return: raw query, s_query, anchors, macros, wordlists
    :rtype: dict
    """

    # get s_query and strip 'within' statement
    s_query = None
    query = query.rstrip(";")
    tokens = query.split(" ")
    if len(tokens) > 2:
        if tokens[-2] == "within":
            query = " ".join(tokens[:-2])
            s_query = tokens[-1]

    # find anchors, macros, wordlists
    anchors = [int(a[1]) for a in find_all(r"@\d", query).keys()]
    macros = list(find_all(r"/[A-Za-z_]+\[\d*\]", query).keys())
    wordlists = list(find_all(r"\$[A-Za-z_]+", query).keys())

    return {
        'query': query,
        's_query': s_query,
        'anchors': anchors,
        'macros': macros,
        'wordlists': wordlists
    }


def decode(text):
    """savely decode a string catching common errors

    """
    try:
        text = text.decode('utf-8')
    except (UnicodeDecodeError, AttributeError):
        text = ""
    return text


#################################
# working on nodes and contexts #
#################################
def _node2cotext(match, matchend, context, contextend):
    """converts the four cpos-values of nodes into a dict of lists of
    cpos, match, and offset

    """
    # get lists
    cpos_list = list(range(context, contextend + 1))
    match_list = [match] * len(cpos_list)
    offset_list = [
        (cpos - match) if cpos < match else
        (cpos - matchend) if cpos > matchend else
        0 for cpos in cpos_list
    ]

    # return object
    result = {
        'match_list': match_list,
        'cpos_list': cpos_list,
        'offset_list': offset_list
    }

    return result


node2cotext = np.vectorize(_node2cotext)


def merge_intervals(inter, start_index=0):
    """for merging contexts; union of intervals. intervals can have
    arbitrary overlaps, but must be sorted by left boundaries.

    :param list inter: list of intervals (each one a pair of left and right boundary)
    :param int start_index: for recursive application

    :return: list of merged intervals
    :rtype: list
    """

    for i in range(start_index, len(inter)-1):
        if inter[i][1] >= inter[i+1][0]:
            new_start = inter[i][0]
            new_end = max(inter[i][1], inter[i+1][1])
            inter[i] = [new_start, new_end]
            del inter[i+1]
            return merge_intervals(inter.copy(), start_index=i)
    return inter


def intersect_intervals(inter_left, inter_right):
    """for merging matches; intersection of intervals. intervals can have
    arbitrary overlaps, but must be sorted by left boundaries.

    :param list inter_left: list of intervals (each one a pair of left and right boundary + ...)
    :param list inter_right: list of intervals (each one a pair of left and right boundary + ...)
    """

    i = j = 0                   # indexes of inter_left and inter_right
    n = len(inter_left)
    m = len(inter_right)

    intervals = list()          # output variable
    while i < n and j < m:

        # compare inter_left[i] with inter_right[j]
        start = max(inter_left[i][0], inter_right[j][0])
        end = min(inter_left[i][1], inter_right[j][1])
        if start <= end:
            intervals.append([start, end] + inter_left[i][2:] + inter_right[j][2:])

        # if inter_left[i]'s right bound is smaller than inter_right[j]'s right bound: increment i
        if inter_left[i][1] < inter_right[j][1]:
            i += 1
        # else increment j
        else:
            j += 1

    return intervals


def calculate_offset(row):
    """ calculates offset of y to x """

    # necessary values
    match_x = row['match_x']
    match_y = row['match_y']

    # if match_y is missing, offset is missing
    if match_y is NA or match_y == -1:
        return NA

    # init matchends if necessary
    if 'matchend_x' in row.keys():
        matchend_x = row['matchend_x']
    else:
        matchend_x = match_x
    if 'matchend_y' in row.keys():
        matchend_y = row['matchend_y']
    else:
        matchend_y = match_y

    # y ... x
    if match_x > matchend_y:
        offset = matchend_y - match_x
    # x ... y
    elif matchend_x < match_y:
        offset = match_y - matchend_x
    # overlap
    else:
        offset = 0

    return offset


##################
# s-att handling #
##################
def merge_s_atts(s_query, s_break, s_meta):
    """ DEPRECATED consistencize s-atts """
    # s_query < s_break < s_meta

    # case 1.1: only s_query
    if s_break is None and s_meta is None:
        s_break = s_meta = s_query

    # case 1.2: only s_break
    elif s_query is None and s_meta is None:
        s_query = s_meta = s_break

    # case 1.3: only s_meta
    elif s_query is None and s_break is None:
        s_query = s_break = s_query

    # case 2.1: s_query and s_break
    elif s_meta is None:
        s_meta = s_break

    # case 2.2: s_query and s_meta
    elif s_break is None:
        s_break = s_meta

    # useless cases
    # case 2.3: s_break and s_meta
    # case 3: all given

    if s_meta is not None and not s_meta.endswith("_id"):
        s_meta = s_meta + "_id"

    logger.info(f"s_query: {s_query} - s_break: {s_break} - s_meta: {s_meta}")
    return s_query, s_break, s_meta


##################
# p-att handling #
##################
def fold_item(item, flags="%cd"):

    if flags is None:
        return item

    # is the item a string or a tuple?
    isstr = isinstance(item, str)
    item = (item, ) if isstr else item

    if "c" in flags:
        # lower-case
        item = [i.lower() for i in item]

    if "d" in flags:
        # TODO align with CWB
        # remove diacritica
        item = [unidecode(i) for i in item]

    item = item[0] if isstr else tuple(item)

    return item


def fold_df(df, flags="%cd"):

    if flags is None:
        return df

    df.index = df.index.map(lambda x: fold_item(x, flags))
    df = df.select_dtypes(include=np.number)
    grouped = df.groupby(df.index)
    df = grouped.aggregate('sum')
    return df


def filter_df(df, path):

    data = pkgutil.get_data(__name__, path)
    items = set(data.decode().split("\n"))
    return df.loc[~df.index.isin(items)]


##########################
# concordance dataframes #
##########################
def dump_left_join(df1, df2, name, drop=True, window=None):
    """join an additional dump df2 to an existing constellation

    :param DataFrame df1: constellation dump === (m, me) ci c ce m_t* me_t* o_t* m_1* me_1* o_1* ... ==
    :param DataFrame df2: additional dump === (m, me) ci c ce ==
    :param str name: name for additional discourseme
    :param bool drop: drop all rows that do not contain all discoursemes within topic context?
    :return: constellation dump including additional dump  === (m, me) ci c ce m_t me_t o_t m_1 me_1 o_1 ... m_name me_name o_name ==
    :rtype: DataFrame
    """

    # merge dumps via contextid ###
    df1 = df1.reset_index()
    df2 = df2.reset_index()[['contextid', 'match', 'matchend']].astype("Int64")
    m = df1.merge(df2, on='contextid', how='left')

    # calculate offset ###
    m['offset_y'] = 0       # init as overlap
    # y .. x
    m.loc[m['match_x'] > m['matchend_y'], 'offset_y'] = m['matchend_y'] - m['match_x']
    # x .. y
    m.loc[m['matchend_x'] < m['match_y'], 'offset_y'] = m['match_y'] - m['matchend_x']
    # missing y
    m.loc[m['match_y'].isna(), 'offset_y'] = NA

    # restrict to complete constellation ###
    if drop:

        m = m.dropna()
        # only keep co-occurrences that are within context
        m = m.loc[(m['matchend_y'] >= m['context']) & (m['match_y'] <= m['contextend'])]

        if window:
            # only keep co-occurrences that are within window
            m = m.loc[abs(m['offset_y']) <= window]

    # rename columns ###
    m = m.rename(columns={'match_x': 'match',
                          'matchend_x': 'matchend',
                          'match_y': 'match_' + name,
                          'matchend_y': 'matchend_' + name,
                          'offset_y': 'offset_' + name})

    # set index ###
    m = m.set_index(['match', 'matchend'])

    return m


def group_lines(df, names):
    """
    convert dataframe:
    === (m, me) ci c ce m0 m0e o0 m1 me1 o1 m2 me2 o2 ===
    with duplicate indices to
    === (m, me) ci c ce m0 m1 m2 ===
    without duplicate indices
    where
    m0 = {(o0, m0, m0e), (o0, m0, m0e), ...}
    m1 = {(o1, m1, m1e), ...}

    """

    # TODO: deduplication necessary?
    df = df.copy()
    df_reduced = df[~df.index.duplicated(keep='first')][['contextid', 'context', 'contextend']]

    for name in names:
        columns = [m + "_" + name for m in ['offset', 'match', 'matchend']]
        df[name] = df[columns].values.tolist()
        df[name] = df[name].apply(tuple)
        df = df.drop(columns, axis=1)
        df_reduced[name] = df.groupby(level=['match', 'matchend'], group_keys=False)[name].apply(set)

    return df_reduced


def aggregate_matches(df, name, context_col='contextid',
                      match_cols=['match', 'matchend']):
    """
    convert dataframe:
    === (m, me) ci ===
    to
    === (ci) {name} {name}_COUNTS {name}_BOOL ===

    """

    # counts
    counts = DataFrame(df[context_col].value_counts()).astype("Int64")
    counts.columns = [name + '_COUNTS']
    counts.index.name = context_col
    counts = counts.sort_values(by=context_col)

    # matches
    matches = df.reset_index()
    matches[name] = matches[match_cols].values.tolist()
    matches[name] = matches[name].apply(tuple)
    matches = matches.groupby('contextid', group_keys=True)[name].apply(set)

    # combine
    table = counts.join(matches)

    # bool
    table[name + '_BOOL'] = (table[name + '_COUNTS'] > 0).fillna(False)

    return table


def format_roles(row, names, s_show, window, htmlify_meta=False):
    """Take a row of a dataframe indexed by match, matchend of the node,
    columns for each discourseme with sets of tuples indicating discourseme positions,
    columns for each s in s_show,
    and a column 'dict' containing the pre-formatted concordance line.

    creates a list (aligned with other lists) of lists of roles; roles are:
    - 'node' (if cpos in index)
    - 'out_of_window' (if offset of cpos from node > window)
    - discourseme names

    :return: concordance line for MMDA frontend
    :rtype: dict

    """

    # TODO directly create relevant objects, no need for frontend to take care of it
    # init
    d = row['dict']
    roles = list()

    # 'out_of_window' | None | 'node'
    role = ['out_of_window' if abs(t) > window else None for t in d['offset']]
    for i in range(d['cpos'].index(row.name[0]), d['cpos'].index(row.name[1]) + 1):
        role[i] = 'node'
    roles.append(role)

    # discourseme names
    for name in names:

        role = [None] * len(d['offset'])
        if row[name] and not isinstance(row[name], float):
            for t in row[name]:

                # check match information
                if len(t) == 2:
                    # lazy definition without offset
                    start = 0
                    end = 1
                elif len(t) == 3:
                    # with offset
                    start = 1
                    end = 2
                else:
                    raise ValueError()

                # skip NAs
                if not isinstance(t[start], int):
                    continue

                # skip the ones too far away
                try:
                    start = d['cpos'].index(t[start])
                    end = d['cpos'].index(t[end]) + 1
                except ValueError:
                    continue

                for i in range(start, end):
                    role[i] = name

        roles.append(role)

    # combine individual roles into one list of lists
    d['role'] = [[a for a in set(r) if a is not None] for r in list(zip(*roles))]

    # add s-attributes
    if htmlify_meta:
        meta = {key: row[key] for key in s_show if not key.endswith("_BOOL")}
        d['meta'] = DataFrame.from_dict(meta, orient='index').to_html(bold_rows=False, header=False, render_links=True)
        for s in s_show:
            if s.endswith("_BOOL"):
                d[s] = row[s]
    else:
        for s in s_show:
            d[s] = row[s]

    return d



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import shlex
import subprocess

from setuptools import Extension, setup

# try:
#     from Cython.Build import cythonize
#     USE_CYTHON = True
# except ImportError:
USE_CYTHON = False              # use cython -2 ccc/cl.pyx instead


########################
# read local resources #
########################

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), mode='rt', encoding='utf-8') as f:
    long_description = f.read()

version = {}
with open(os.path.join(here, 'ccc', 'version.py'), mode='rt', encoding='utf-8') as f:
    exec(f.read(), version)


##############
# cwb-config #
##############

# - version
cwb_version = subprocess.run(shlex.split("cwb-config -v"), capture_output=True).stdout.decode().strip()
# - cqp binaries
cwb_bindir = subprocess.run(shlex.split("cwb-config --bindir"), capture_output=True).stdout.decode().strip()
# - effective registry directory or directories:
cwb_registry = subprocess.run(shlex.split("cwb-config -r"), capture_output=True).stdout.decode().strip()
# - libdir (CL library)
cwb_libdir = subprocess.run(shlex.split("cwb-config --libdir"), capture_output=True).stdout.decode().strip()
# - incdir (C header)
cwb_incdir = subprocess.run(shlex.split("cwb-config --incdir"), capture_output=True).stdout.decode().strip()
# - compiler flags for linking against CL library
cwb_compiler_flags = subprocess.run(shlex.split("cwb-config -I"), capture_output=True).stdout.decode().strip()
# - linker flags for linking against CL library
cwb_linker_flags = subprocess.run(shlex.split("cwb-config -L"), capture_output=True).stdout.decode().strip()


####################################
# define (and compile) C-extension #
####################################

# ensure compatibility with CWB v3.4.36 and below
if int(cwb_version.split(".")[0]) == 3 and int(cwb_version.split(".")[1]) == 4 and int(cwb_version.split(".")[2]) < 37:
    cwb_linker_flags = "-L/usr/local/lib -lcl  -lm   -lpcre -lglib-2.0"

# define include directories, library directories, and library names
libraries = [t[2:] for t in shlex.split(cwb_linker_flags) if t.startswith("-l")]
inc_dirs = [cwb_incdir] + [t[2:] for t in shlex.split(cwb_compiler_flags) if t.startswith("-I")]
lib_dirs = [cwb_libdir] + [t[2:] for t in shlex.split(cwb_linker_flags) if t.startswith("-L")]

ccc_cl = Extension(
    name="ccc.cl",
    sources=['ccc/cl' + ('.pyx' if USE_CYTHON else '.c')],
    include_dirs=inc_dirs,  # list of directories to search for C/C++ header files
    library_dirs=lib_dirs,  # list of directories to search for C/C++ libraries at link time
    libraries=libraries     # list of library names (not filenames or paths) to link against
)

# cythonize?
extensions = [ccc_cl]
if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions)


#################
# actual set-up #
#################
setup(
    name="cwb-ccc",
    version=version["__version__"],
    description="CWB wrapper to extract concordances and score frequency lists",
    license='GNU General Public License v3 or later (GPLv3+)',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Philipp Heinrich",
    author_email="philipp.heinrich@fau.de",
    url="https://github.com/ausgerechnet/cwb-ccc",
    packages=[
        'ccc'
    ],
    ext_modules=extensions,
    python_requires='>=3.8.0',
    install_requires=[
        "numpy>=1.24.0,<2.0",
        "numexpr>=2.8.6,<3.0",
        "bottleneck==1.4.0",
        "pandas>=2.0,<3.0",
        "association-measures>=0.3.0,<0.4",
        "unidecode>=1.3.8,<2.0",
        "pyyaml>=6.0.1,<7.0",
        "trieregex>=1.0.0,<1.1"
    ],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Development Status :: 4 - Beta",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Cython'
    ],
)



import pytest

from ccc import Corpus
from ccc.counts import score_counts
from ccc.keywords import Keywords, keywords

from .conftest import DATA_PATH


def get_corpus(corpus_settings, data_dir=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_dir=corpus_settings['registry_dir'],
        lib_dir=corpus_settings.get('lib_dir', None),
        data_dir=data_dir
    )


@pytest.mark.meta
def test_keywords_from_satt_values(germaparl):

    corpus = get_corpus(germaparl)
    df_dump = corpus.query_s_att(
        "text_party", {"GRUENE", "Bündnis 90/Die Grünen"}
    ).df

    # keywords
    keywords = Keywords(corpus, df_dump, p_query='lemma')
    lines = keywords.show(order='log_ratio', cut_off=None)
    assert lines.index[0] == "Oppositionsfraktion"


@pytest.mark.query
def test_keywords_from_query(germaparl):

    # get subcorpus as dump
    corpus = get_corpus(germaparl)
    df_dump = corpus.query('"Seehofer" expand to s').df

    # keywords
    keywords = Keywords(corpus, df_dump)
    lines = keywords.show(order='log_ratio')
    assert lines.index[1] == "Gesundheitsreform"


@pytest.mark.query
def test_keywords_from_dump(germaparl):

    # get subcorpus as dump
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query('"und" expand to s')

    # keywords
    keywords = Keywords(corpus, df_dump)
    lines = keywords.show(order='log_likelihood', min_freq=10)
    assert lines.index[1] == "Dame"


@pytest.mark.subcorpus
def test_keywords_switch(germaparl):

    # get some regions
    corpus = get_corpus(germaparl)
    df_all = corpus.query('"und" expand to s', name='Und_all').df

    df_head = df_all.head(500)
    df_tail = df_all.tail(500)

    # will show keywords for head
    keywords = Keywords(corpus, df_dump=df_head, p_query="lemma")
    lines_head = keywords.show(order='log_likelihood')

    # will show keywords for tail
    keywords = Keywords(corpus, df_dump=df_tail, p_query="lemma")
    lines_tail = keywords.show(order='log_likelihood')

    assert not lines_head.equals(lines_tail)


@pytest.mark.query
def test_keywords_combo(germaparl):

    # get subcorpus as dump
    corpus = get_corpus(germaparl)
    dump = corpus.query('"und" expand to s')

    # keywords
    lines = dump.keywords(["lemma", "pos"], order='log_likelihood', min_freq=10)
    assert lines.index[0] == "und KON"


def test_keywords(germaparl):

    corpus = get_corpus(germaparl)

    left = corpus.marginals(p_atts=['lemma', 'pos'])[['freq']].rename(
        columns={'freq': 'f1'}
    )
    right = corpus.marginals(p_atts=['lemma', 'pos'])[['freq']].rename(
        columns={'freq': 'f2'}
    )

    df = left.join(right).fillna(0)  # , downcast='infer')
    df['N1'] = df['f1'].sum()
    df['N2'] = df['f2'].sum()
    kw = score_counts(df, order='O11')
    assert kw.iloc[0]['O11'] == 11469
    assert kw.iloc[0]['conservative_log_ratio'] == 0


@pytest.mark.subcorpus
def test_keywords_func(germaparl):

    corpus = get_corpus(germaparl)
    green = corpus.query_s_att("text_party", {"GRUENE", "Bündnis 90/Die Grünen"})
    red = corpus.query_s_att("text_party", {"SPD", "S.P.D."})
    kw = keywords(green, red, ['lemma'], ['lemma'], 'conservative_log_ratio')
    assert kw.index[0] == 'Dr.'


@pytest.mark.subcorpus
def test_keywords_func_corpora(germaparl):

    corpus = get_corpus(germaparl)
    kw = keywords(corpus, corpus, ['lemma'], ['word'], 'conservative_log_ratio')
    assert kw.index[0] == 'sie'


@pytest.mark.subcorpus
def test_keywords_func_corpora_p_atts(germaparl):

    corpus = get_corpus(germaparl)
    kw = keywords(corpus, corpus, ['lemma', 'pos'], ['lemma', 'pos'], 'conservative_log_ratio')

    assert kw.index[0] == '! $.'


@pytest.mark.subcorpus
def test_keywords_func_p_atts(germaparl):

    corpus = get_corpus(germaparl)
    green = corpus.query_s_att("text_party", {"GRUENE", "Bündnis 90/Die Grünen"})
    red = corpus.query_s_att("text_party", {"SPD", "S.P.D."})
    kw = keywords(green, red, ['lemma', 'pos'], ['lemma', 'pos'], 'conservative_log_ratio')

    assert kw.index[0] == 'Dr. NN'


@pytest.mark.subcorpus
def test_keywords_func_p_atts_2(germaparl):

    corpus = get_corpus(germaparl)
    green = corpus.query_s_att("text_party", {"GRUENE", "Bündnis 90/Die Grünen"})
    red = corpus.query_s_att("text_party", {"SPD", "S.P.D."})
    kw = keywords(green, red, ['word', 'pos'], ['lemma', 'pos'], 'conservative_log_ratio')

    assert kw.index[0] == 'der ART'



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""cqpy.py

read-write support for CQPY query files

- a CQPY file contains a YAML header and the actual CQP input
- YAML header *can* be commented
- query formatting (whitespace incl. linebreaks) is preserved

"""
import json
import logging
import re

# requirements
import yaml

logger = logging.getLogger(__name__)


def cqpy_load_json(query_path):
    """ DEPRECATED: support for query.json files pre 2020-08-29 """

    # read file
    with open(query_path, "rt") as f:
        query = json.loads(f.read())

    out = dict()
    # needs following keys:
    # 'meta'
    # 'corpus'
    # 'display'
    # 'anchors'
    # 'cqp'

    # we defintinely need this one
    out['cqp'] = query['query']

    # meta
    if 'name' not in query.keys():
        query['name'] = query_path.split("/")[-1].split(".")[0]
    if 'pattern' not in query.keys():
        query['pattern'] = 9999
    out['meta'] = {
        'name': query['name'],
        'pattern': int(query['pattern'])
    }

    # corpus
    if 'corpus_name' not in query.keys():
        query['corpus_name'] = "BREXIT_V20190522_DEDUP"
    if 'lib_dir' not in query.keys():
        query['lib_dir'] = None
    out['corpus'] = {
        'corpus_name': query['corpus_name'],
        'lib_dir': query['lib_dir']
    }

    # display
    if 's_show' not in query.keys():
        query['s_show'] = ['tweet_id']
    if 'p_show' not in query.keys():
        query['p_show'] = ['word', 'lemma']
    if 'p_text' not in query.keys():
        query['p_text'] = 'word'
    if 'p_slots' not in query.keys():
        query['p_slots'] = 'lemma'
    out['display'] = {
        'p_show': query['p_show'],
        's_show': query['s_show'],
        'p_slots': query['p_slots'],
        'p_text': query['p_text']
    }

    # anchors
    out['anchors'] = {
        'corrections': dict(),
        'slots': dict()
    }
    if 'anchors' in query.keys():
        anchors = query.pop('anchors')
        for anchor, offset, idx, clear in anchors:
            if offset != 0:
                out['anchors']['corrections'][anchor] = offset
            if idx is not None:
                out['anchors']['slots'][idx] = anchor
    if 'regions' in query.keys():
        for start, end, idx, clear in query['regions']:
            if idx is not None:
                out['anchors']['slots'][idx] = [start, end]

    # query
    if 's_context' not in query.keys():
        query['s_context'] = 'tweet'
    if 'context' not in query.keys():
        query['context'] = None
    out['query'] = {
        'context': query['context'],
        's_context': query['s_context']
    }

    return out


def cqpy_load(path):
    """
    load cqpy query file from path, return dictionary
    """

    return cqpy_loads(open(path, 'rt').read())


def cqpy_loads(doc):
    """
    load cqpy query file from string, return dictionary
    """

    try:
        # commented YAML header
        header, cqp = doc.split("# ---\n")
    except ValueError:
        # uncommented YAML header
        header, cqp = doc.split("---\n")
    except yaml.scanner.ScannerError:
        logger.error("not a valid CQPY file")
        return dict()

    # read header
    header_list = [re.sub(r"\A# ", "", h) for h in header.split("\n")]
    query = yaml.load("\n".join(header_list), Loader=yaml.FullLoader)

    # add CQP
    query['cqp'] = cqp.lstrip().rstrip()

    # post-process anchors
    query = check_anchors(query)

    return query


def check_anchors(query):
    """
    make sure integer anchors are indeed integers
    """
    if 'anchors' in query:
        if 'corrections' in query['anchors']:
            corrections_int = dict()
            for k, c in query['anchors']['corrections'].items():
                try:
                    corrections_int[int(k)] = c
                except ValueError:  # for 'match', 'matchend', etc.
                    pass
            query['anchors']['corrections'] = corrections_int
    return query


def cqpy_dump(query, path, comment=True, mode='wt'):
    """
    write query in CQPY format to path
    """

    open(path, mode=mode).write(cqpy_dumps(query, comment))


def cqpy_dumps(query, comment=True):
    """
    serialize query as CQPY string
    """

    # get actual query
    query = query.copy()
    cqp = query.pop('cqp')

    # header
    out = "--- # CQPY query file\n"
    out += yaml.dump(query, default_flow_style=False)
    out += "---"
    # comment header?
    if comment:
        out = "\n".join(["# " + line for line in out.split("\n")])
    # actual cqp query
    out += "\n\n" + cqp + "\n"

    return out


def run_query(corpus, query,
              context=None, context_break=None, match_strategy='longest',
              corrections={}, slots={},
              p_show=['word', 'lemma'], s_show=[], cut_off=None, form='slots'):
    """Execute a query in a corpus. Return concordance lines as dataframe
    in 'slots' format. Parameters that are not set in the query-dict
    will be overwritten by method arguments.

    :param Corpus corpus: ccc.Corpus
    :param dict query: query with the following parts:
                       - cqp
                       - query > context        None
                       - query > context_break  None
                       - query > match_strategy "longest"
                       - anchors > corrections  {}
                       - anchors > slots        {}
                       - display > p_show       ['word', 'lemma']
                       - display > s_show       []
                       - display > cut_off      None
                       - display > form         'slots'

    :return: concordances lines
    :rtype: DataFrame

    """

    # the actual CQP query
    cqp = query['cqp']

    # determine query parameters
    if 'query' in query:

        # backwards compatability
        if 's_context' in query['query']:
            logger.warning("use of 's_context' is deprecated")
            query['query']['context_break'] = query['query'].pop('s_context')

        context = query['query'].get('context', context)
        context_break = query['query'].get('context_break', context_break)
        match_strategy = query['query'].get('match_strategy', match_strategy)

    # determine anchor parameters
    if 'anchors' in query:
        query = check_anchors(query)
        corrections = query['anchors'].get('corrections', corrections)
        slots = query['anchors'].get('slots', slots)

    # determine display parameters
    if 'display' in query:

        p_show = query['display'].get('p_show', p_show)
        s_show = query['display'].get('s_show', s_show)
        cut_off = query['display'].get('cut_off', cut_off)
        form = query['display'].get('form', form)

        # backwards compatability
        for p in ['p_slots', 'p_text']:
            if p in query['display']:
                # logger.warning(f"use of '{p}' is deprecated")
                if query['display'][p] not in p_show:
                    p_show += [query['display'][p]]

    # query the corpus
    dump = corpus.query_cqp(
        cqp_query=cqp,
        context=context,
        context_break=context_break,
        corrections=corrections,
        match_strategy=match_strategy,
    )

    # retrieve concordance lines
    lines = dump.concordance(
        p_show=p_show,
        s_show=s_show,
        slots=slots,
        cut_off=cut_off,
        form=form
    )

    # post-process: only return relevant columns
    if 'display' in query:
        if 'p_text' in query['display']:
            drop = [p for p in p_show if p != query['display']['p_text']]
            lines = lines.drop(drop, axis=1)
        if 'p_slots' in query['display']:
            drop = list()
            for slot in slots.keys():
                drop = ["_".join([slot, p]) for p in p_show
                        if p != query['display']['p_slots']]
                lines = lines.drop(drop, axis=1)

    return lines



from ccc import Corpus
from ccc.utils import (filter_df, fold_item, intersect_intervals,
                       merge_intervals, preprocess_query)

from .conftest import DATA_PATH


def test_preprocess_query():
    query = (
        "/nobody[] <vp>[]*@0:[lemma = $verbs_cog_positive] "
        "[]* @1:[]</vp> <np>@2:[::][]* [lemma = $adj_cog_negative]* "
        "[lemma = $nouns_person_negative][]* @3:[]</np> "
        "within tweet;"
    )
    query = preprocess_query(query)
    assert query['anchors'] == [0, 1, 2, 3]
    assert query['s_query'] == 'tweet'
    assert query['wordlists'] == ["$verbs_cog_positive", "$adj_cog_negative",
                                  "$nouns_person_negative"]
    assert query['macros'] == ['/nobody[]']

    query = (
        "<vp>[]*@0:[lemma = $verbs_cog_positive] "
        "[]* @1:[]</vp> <np>@2:[::][]* [lemma = $adj_cog_negative]* "
        "[lemma = $nouns_person_negative][]* @3:[]</np>;"
    )
    query = preprocess_query(query)
    assert query['s_query'] is None
    assert query['macros'] == []


def test_preprocess_anchor_query():
    query = (
        '@0[lemma="Angela"]? @1[lemma="Merkel"] '
        '[word="\\("] @2[lemma="CDU"] [word="\\)"]'
    )
    query_1 = preprocess_query(query)

    query = (
        '@0[lemma="Angela"]? @1[lemma="Merkel"] '
        '[word="\\("] @2[lemma="CDU"] [word="\\)"];'
    )
    query_2 = preprocess_query(query)

    query = (
        '@0[lemma="Angela"]? @1[lemma="Merkel"] '
        '[word="\\("] @2[lemma="CDU"] [word="\\)"] within tweet;'
    )
    query_3 = preprocess_query(query)

    query = (
        '@0[lemma="Angela"]? @1[lemma="Merkel"] '
        '[word="\\("] @2[lemma="CDU"] [word="\\)"] within s'
    )
    query_4 = preprocess_query(query)

    assert query_1['s_query'] == query_2['s_query'] is None
    assert query_3['s_query'] == 'tweet'
    assert query_4['s_query'] == 's'
    assert query_1['query'] == query_2['query'] == query_3['query'] == query_4['query']
    assert query_1['anchors'] == query_2['anchors'] == query_3['anchors'] == query_4['anchors']


def test_fold_items():
    items_aber = ['aber', 'äber', 'AbEr', 'ÄBER']
    items_francais = ['français', 'Fráncàis']
    assert len(set([fold_item(item) for item in items_aber])) == 1
    assert len(set([fold_item(item) for item in items_francais])) == 1


def test_filter_df(germaparl):
    c = Corpus(germaparl['corpus_name'], registry_dir=germaparl['registry_dir'], data_dir=DATA_PATH)
    subcorpus = c.query(germaparl['query'])
    coll = subcorpus.collocates()
    assert ',' in coll.index
    coll_filtered = filter_df(coll, 'resources/stopwords-de.txt')
    assert ',' not in coll_filtered.index


def test_merge_intervals():

    intervals = [[1, 3], [2, 4], [5, 9], [6, 10]]
    merge = merge_intervals(intervals)
    assert merge == [[1, 4], [5, 10]]

    intervals = [[0, 5], [2, 4], [6, 9], [6, 10]]
    merge = merge_intervals(intervals)
    assert merge == [[0, 5], [6, 10]]


def test_intersect_intervals():

    intervals_left = [[0, 1801], [7438, 7449]]
    intervals_right = [[458, 463], [548, 553], [1025, 1500], [2540, 2670], [7438, 7439], [7440, 7449], [8500, 8600]]
    merge = intersect_intervals(intervals_left, intervals_right)
    assert merge == [[458, 463], [548, 553], [1025, 1500], [7438, 7439], [7440, 7449]]

    intervals_left = [[0, 1801], [7438, 7449]]
    intervals_right = [[458, 463], [548, 553], [1025, 1500], [1800, 1802], [7438, 7439], [7440, 7449], [8500, 8600]]
    merge = intersect_intervals(intervals_left, intervals_right)
    assert merge == [[458, 463], [548, 553], [1025, 1500], [1800, 1801], [7438, 7439], [7440, 7449]]



import pandas as pd
import pytest

from ccc import Corpus
from ccc.concordances import Concordance

from .conftest import DATA_PATH


def get_corpus(corpus_settings, data_dir=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_dir=corpus_settings['registry_dir'],
        lib_dir=corpus_settings.get('lib_dir', None),
        data_dir=data_dir
    )


###############
# CONCORDANCE #
###############
@pytest.mark.concordance
@pytest.mark.quick_conc
def test_ccc_quick_conc(germaparl):

    corpus = get_corpus(germaparl)

    topic_query = '[lemma="die"]'
    s_context = 's'
    window = 5
    filter_queries = highlight_queries = {
        'CSU': '[lemma="CSU"]',
        'CDU': '[lemma="CDU"]'
    }

    conc = corpus.quick_conc(
        topic_query=topic_query,
        filter_queries=filter_queries,
        highlight_queries=highlight_queries,
        window=window,
        s_context=s_context,
        cut_off=100,
        order='first'
    )

    assert len(conc) == 100
    assert 'cpos' in conc[0]


@pytest.mark.concordance
@pytest.mark.quick_conc
def test_ccc_quick_conc_2(germaparl):

    corpus = get_corpus(germaparl)

    s_context = 's'
    queries = {
        'Herr': '[lemma="Herr"]',
        'CSU': '[lemma="CSU"]',
        'CDU': '[lemma="CDU"]'
    }

    conc = corpus.quick_conc(
        s_context=s_context,
        topic_query={},
        window=0,
        filter_queries=queries,
        cut_off=100,
        order='first'
    )

    assert len(conc) == 100
    assert 'cpos' in conc[0]


@pytest.mark.concordance
@pytest.mark.quick_conc
def test_ccc_quick_conc_random(germaparl):

    corpus = get_corpus(germaparl)

    topic_query = '[lemma="die"]'
    s_context = 's'
    window = 5
    filter_queries = highlight_queries = {
        'CSU': '[lemma="CSU"]',
        'CDU': '[lemma="CDU"]'
    }

    conc = corpus.quick_conc(
        topic_query=topic_query,
        filter_queries=filter_queries,
        highlight_queries=highlight_queries,
        window=window,
        s_context=s_context,
        cut_off=100,
        order=50
    )

    assert len(conc) == 100
    assert 'cpos' in conc[0]


@pytest.mark.raw
def test_concordance_simple(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.query('"CSU"').df
    concordance = Concordance(corpus, df_dump)
    lines = concordance.simple(
        df_dump,
        p_show=['word', 'lemma']
    )
    assert len(lines) == len(df_dump)
    assert all(col in lines.columns for col in [
        'word', 'lemma'
    ])


@pytest.mark.raw
def test_concordance_simple_nocontext(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.query('[lemma="gehen"]', context=None).df
    concordance = Concordance(corpus, df_dump)
    lines = concordance.simple(
        df_dump,
        p_show=['word']
    )
    assert len(lines) == len(df_dump)
    assert all(col in lines.columns for col in [
        'word'
    ])


@pytest.mark.raw
def test_concordance_kwic(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.query('[lemma="gehen"]').df
    concordance = Concordance(corpus, df_dump)
    lines = concordance.kwic(
        df_dump,
        p_show=['word', 'lemma']
    )
    assert len(lines) == len(df_dump)
    assert all(col in lines.columns for col in [
        'left_word', 'node_word', 'right_word',
        'left_lemma', 'node_lemma', 'right_lemma'
    ])


@pytest.mark.raw
def test_concordance_kwic_nocontext(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.query('[lemma="gehen"]', context=None).df
    concordance = Concordance(corpus, df_dump)
    lines = concordance.kwic(
        df_dump,
        p_show=['word']
    )
    assert len(lines) == len(df_dump)
    assert all(col in lines.columns for col in [
        'left_word', 'node_word', 'right_word'
    ])


@pytest.mark.raw
def test_concordance_slots_singletons(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        r'[pos="NE"]? @1[pos="NE"] @2"\[" ([word="[A-Z]+"]+ "/"?)+ @3"\]"'
    )
    df_dump = corpus.query(query, context=2, context_break='s',
                           match_strategy='longest', corrections={2: +1, 3: -1}).df
    concordance = Concordance(corpus, df_dump)
    lines = concordance.slots(
        df_dump,
        ['word', 'lemma']
    )
    assert set(lines.columns) == {
        "word", "lemma",
        "1_word", "1_lemma", "2_word", "2_lemma", "3_word", "3_lemma",
        "match..matchend_word", "match..matchend_lemma"
    }


@pytest.mark.raw
def test_concordance_slots_regions(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        r'[pos="NE"]? @1[pos="NE"] @2"\[" ([word="[A-Z]+"]+ "/"?)+ @3"\]"'
    )
    df_dump = corpus.query(query, context=10, context_break='s',
                           match_strategy='longest', corrections={2: +1, 3: -1}).df
    concordance = Concordance(corpus, df_dump)
    lines = concordance.slots(
        df_dump,
        ['word'],
        slots=[['match', 1], [2, 3]]
    )
    assert set(lines.columns) == {
        "word", "match..1_word", "2..3_word"
    }


@pytest.mark.raw
def test_concordance_slots_regions_dict(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        r'[pos="NE"]? @1[pos="NE"] @2"\[" ([word="[A-Z]+"]+ "/"?)+ @3"\]"'
    )
    df_dump = corpus.query(query, context=10, context_break='s',
                           match_strategy='longest', corrections={2: +1, 3: -1}).df
    concordance = Concordance(corpus, df_dump)
    lines = concordance.slots(
        df_dump,
        ['word'],
        slots={'mp': ['match', 1], 'party': [2, 3]}
    )
    assert set(lines.columns) == {
        'word', 'mp_word', 'party_word'
    }


@pytest.mark.line
def test_concordance_export_dict(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\["] @1[lemma="CDU"] "/" "CSU" [word="\\]"]'
    )
    result = corpus.query(query)
    concordance = Concordance(corpus, result.df)
    line = result.df.iloc[0]
    text_line = concordance._export(
        line.name, line, p_show=['word', 'pos'], form='dict'
    )
    assert isinstance(text_line, dict)
    assert 'cpos' in text_line


@pytest.mark.line
def test_concordance_export_dataframe(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\["] @1[lemma="CDU"] "/" "CSU" [word="\\]"]'
    )
    result = corpus.query(query, context_break='s')
    concordance = Concordance(corpus, result.df)
    line = result.df.iloc[0]
    text_line = concordance._export(
        line.name, line, p_show=['word', 'pos'], form='dataframe'
    )
    assert isinstance(text_line, pd.DataFrame)


@pytest.mark.raw
def test_concordance_dict(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.query('[lemma="gehen"]', context=None).df
    concordance = Concordance(corpus, df_dump)
    lines = concordance.dict(
        df_dump,
        p_show=['word']
    )
    assert isinstance(lines['dict'].iloc[0], dict)
    assert 'word' in lines['dict'].iloc[0]


@pytest.mark.raw
def test_concordance_dataframes(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\["] @1[lemma="CDU"] "/" "CSU" [word="\\]"]'
    )
    result = corpus.query(query, context_break='s',
                          match_strategy='longest')
    concordance = Concordance(corpus, result.df)
    df = concordance.dict(result.df, p_show=['word', 'lemma'])
    lines = concordance.dataframe(df, p_show=['word', 'lemma'])
    assert 'dataframe' in lines.columns
    assert isinstance(lines['dataframe'].iloc[0], pd.DataFrame)


@pytest.mark.lines
def test_concordance_lines(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\["] @1[lemma="CDU"] "/" "CSU" [word="\\]"]'
    )
    result = corpus.query(query, context_break='s',
                          match_strategy='longest')

    concordance = Concordance(corpus, result.df)

    # default = simple
    lines = concordance.lines()
    assert len(lines) > 10
    assert 'word' in lines.columns

    # kwic
    lines = concordance.lines(form='kwic', cut_off=10)
    assert all(
        elem in lines.columns for elem in ['left_word', 'node_word', 'right_word']
    )
    assert len(lines) == 10

    # kwic with s-attribute
    lines = concordance.lines(form='kwic', s_show=['text_id'], cut_off=10)
    assert len(lines) == 10
    assert 'text_id' in lines.columns

    # slots
    lines = concordance.lines(form='slots', s_show=['text_id'], cut_off=10)
    assert len(lines) == 10
    assert all(
        elem in lines.columns for elem in ['match..matchend_word', '1_word']
    )

    # dict
    lines = concordance.lines(form='dict', s_show=['text_id'], cut_off=10)
    assert len(lines) == 10
    assert all(
        elem in lines.columns for elem in ['dict', 'text_id']
    )

    # dict
    lines = concordance.lines(form='dataframe', s_show=['text_id'], cut_off=10)
    assert len(lines) == 10
    assert all(
        elem in lines.columns for elem in ['dataframe', 'text_id']
    )


@pytest.mark.lines
def test_concordance_many(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[lemma="oder"]'
    )
    result = corpus.query(query)
    concordance = Concordance(corpus, result.df)
    lines = concordance.lines()
    assert len(lines) == 100


@pytest.mark.lines
def test_concordance_p_atts(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\["] [lemma="CDU"] "/" ".*" [word="\\]"]'
    )
    result = corpus.query(query, context_break='s')
    concordance = Concordance(corpus, result.df)
    lines = concordance.lines(p_show=['lemma', 'pos'], form='dataframe')
    assert 'pos' in lines.iloc[0]['dataframe'].columns
    assert len(lines) == 13


@pytest.mark.lines
def test_concordance_anchors(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\["] @1[lemma="CDU"] "/" @2".*" [word="\\]"]'
    )
    result = corpus.query(query, context_break='s')
    concordance = Concordance(corpus, result.df)
    lines = concordance.lines(p_show=['lemma', 'pos'], form='dataframe')
    assert len(lines) == 13


@pytest.mark.lines
def test_concordance_anchors_weird(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\["] @9[lemma="CDU"] "/" @2".*" @5[word="\\]"]'
    )
    result = corpus.query(query, context_break='s')
    concordance = Concordance(corpus, result.df)
    lines = concordance.lines(order='random', cut_off=100, form='dataframe')
    assert len(lines) == 13


@pytest.mark.lines
def test_concordance_form_simple(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\["] @1[lemma="CDU"] "/" @2".*" [word="\\]"]'
    )
    result = corpus.query(query, context_break='s')
    concordance = Concordance(corpus, result.df)
    lines = concordance.lines(order='random', cut_off=100, form='simple')
    assert len(lines) == 13


@pytest.mark.lines
def test_concordance_form_kwic(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\["] @1[lemma="CDU"] "/" @2".*" [word="\\]"]'
    )
    result = corpus.query(query, context_break='s')
    concordance = Concordance(corpus, result.df)
    lines = concordance.lines(order='random', cut_off=100, form='kwic')
    assert len(lines) == 13


@pytest.mark.lines
def test_concordance_persistence(germaparl):
    corpus = get_corpus(germaparl)
    query_1 = (
        '[word="\\["] @1[lemma="CDU"] "/" @2".*" [word="\\]"]'
    )
    query_2 = (
        '"und"'
    )

    # will show results for query_1
    dump = corpus.query(query_1, context_break='s')
    concordance = Concordance(corpus, dump.df)
    line_1 = concordance.lines(cut_off=1, form='dataframe')
    df_1 = line_1['dataframe'].iloc[0]

    # will show results for query_1
    dump = corpus.query(query_2, context_break='s')
    line_2 = concordance.lines(cut_off=1, form='dataframe')
    df_2 = line_2['dataframe'].iloc[0]

    # will show results for query_2
    concordance = Concordance(corpus, dump.df)
    line_3 = concordance.lines(cut_off=1, form='dataframe')
    df_3 = line_3['dataframe'].iloc[0]

    assert df_1.equals(df_2)
    assert not df_2.equals(df_3)


@pytest.mark.fail
def test_concordance_empty(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[lemma="Gerhard"]? [lemma="NAHH"]'
    )
    dump = corpus.query(query)
    conc = Concordance(corpus, dump.df)
    assert conc.lines().empty


@pytest.mark.fail
def test_concordance_format(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[lemma="Gerhard"]? [lemma="Schröder"]'
    )
    dump = corpus.query(query)
    conc = Concordance(corpus, dump.df)
    conc.lines(form='fail')


@pytest.mark.fail
def test_concordance_order(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[lemma="Gerhard"]? [lemma="Schröder"]'
    )
    dump = corpus.query(query)
    conc = Concordance(corpus, dump.df)
    conc.lines(order='fail')


def test_concordance_last(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[lemma="Gerhard"]? [lemma="Schröder"]'
    )
    dump = corpus.query(query)
    conc = Concordance(corpus, dump.df)
    assert isinstance(conc.lines(order='last'), pd.DataFrame)


def test_concordance_fallback(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[lemma="Gerhard"]? [lemma="Schröder"]'
    )
    dump = corpus.query(query)
    conc = Concordance(corpus, dump.df)
    assert isinstance(conc.lines(order='last', form='simple', p_show=['word', 'lemma']), pd.DataFrame)


@pytest.mark.benchmark
def test_perf_concordance(benchmark, germaparl):
    benchmark.pedantic(test_concordance_dataframes, kwargs={'germaparl': germaparl}, rounds=10, iterations=5)
